#!/usr/bin/env python3

"""Plots outermost contours and max-speed contours.

Red for anti-cyclones, blue for cyclones. 

Without the --light option. Empty circles for extrema
with a valid outermost contour, empty squares for other
extrema. Squares on outermost contour for a well-defined but invalid
outermost contour. Crosses on outermost contour for a valid outermost
contour but with no distinct max-speed contour. Filled circles on
outermost contour and max-speed contour for a valid outermost contour
with a distinct max-speed contour.

With the --light option. Crosses for extrema with a valid outermost
contour.

This script takes about 30 s of CPU for a 90° by 90° window, about 4
mn for a 180° by 180° window.

"""

import shapefile
import numpy as np
import cartopy.crs as ccrs
from os import path
from matplotlib import patches
import f90nml
import sys
import util_eddies

def select_ishapes(d, handler, window = None):
    """Select ishapes at date d and in window."""
    
    assert d >= handler["d_init"]
    
    if d == handler["d_init"]:
        ishape_first = 0
    else:
        # d > handler["d_init"]
        ishape_first = handler["ishape_last"][d - handler["d_init"] - 1] + 1

    if window is None:
        ishape_list = list(range(ishape_first,
                                 handler["ishape_last"][d - handler["d_init"]]
                                 + 1))
    else:
        ishape_list = []

        for ishape in range(ishape_first,
                            handler["ishape_last"][d - handler["d_init"]] + 1):
            shape_rec_extr = handler["readers"]["extremum"].shapeRecord(ishape)
            points = shape_rec_extr.shape.points[0]

            if util_eddies.in_window(points, window):
                ishape_list.append(ishape)

    return ishape_list

def snapshot(ax, ishape_list, handler, *, dashed = False, light = False,
             src_crs = ccrs.PlateCarree()):
    """Plots extrema, outermost contours and max-speed contours from a
    given SHPC.

    dashed: boolean. ax should be an instance of GeoAxes.

    """
    
    if handler["cyclones"]:
        color_outer = "blue"
        color_m_s = "cyan"
    else:
        color_outer = "red"
        color_m_s = "magenta"

    for ishape in ishape_list:
        shape_rec_extr = handler["readers"]["extremum"].shapeRecord(ishape)
        shape_outer = handler["readers"]["outermost_contour"].shape(ishape)

        if "max_speed_contour" in handler["readers"]:
            shape_m_s = handler["readers"]["max_speed_contour"].shape(ishape)
        else:
            shape_m_s = None

        points = shape_rec_extr.shape.points[0]
        lines = ax.plot(points[0], points[1], markersize = 10,
                        color = color_outer, fillstyle = "none",
                        transform = src_crs)

        if not hasattr(shape_rec_extr.record, "valid")  \
           or shape_rec_extr.record.valid == 1:
            if light:
                lines[0].set_marker("+")
            else:
                lines[0].set_marker("o")
        elif not light:
            # Invalid outermost contour
            lines[0].set_marker("s")

        if not light:
            ax.annotate(str(shape_rec_extr.record.eddy_index),
                        ax.projection.transform_point(points[0], points[1],
                                                      src_crs),
                        xytext = (3, 3), textcoords = "offset points")

        if shape_outer.shapeType != shapefile.NULL and not light \
           or shape_rec_extr.record.valid:
            points = np.array(shape_outer.points)
            lines = ax.plot(points[:, 0], points[:, 1], color = color_outer,
                            transform = src_crs)

            if not light:
                if hasattr(shape_rec_extr.record, "valid")  \
                   and shape_rec_extr.record.valid == 0:
                    # Invalid outermost contour
                    lines[0].set_marker("s")
                    lines[0].set_fillstyle("none")
                elif shape_m_s == None \
                     or shape_m_s.shapeType == shapefile.NULL:
                    lines[0].set_marker("x")
                else:
                    lines[0].set_marker("o")

            if dashed: lines[0].set_linestyle("dashed")

            if shape_m_s != None and shape_m_s.shapeType != shapefile.NULL:
                points = np.array(shape_m_s.points)
                lines = ax.plot(points[:, 0], points[:, 1], color = color_m_s,
                                transform = src_crs)
                if not light: lines[0].set_marker("o")
                if dashed: lines[0].set_linestyle("dashed")

def plot_grid_bb(grid_nml, ax):
    """Grid bounding box."""

    if grid_nml is not None:
        step = grid_nml["STEP_DEG"]
        rect = patches.Rectangle(grid_nml["corner_deg"],
                                 (grid_nml["nlon"] - 1) * step[0],
                                 (grid_nml["nlat"] - 1) * step[1],
                                 edgecolor="black", fill=False)
        ax.add_patch(rect)

if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import argparse
    import netCDF4

    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--date", type = int, help = "date index")
    parser.add_argument("-g", "--grid", help = "plot grid",
                        action = "store_true")
    parser.add_argument("-w", "--window", help = "choose a limited plot window",
                        type = float, nargs = 4,
                        metavar = ("llcrnrlon", "llcrnrlat", "urcrnrlon",
                                   "urcrnrlat"))
    parser.add_argument("-l", "--light", help = "lighter plot",
                        action = "store_true")
    parser.add_argument("--dashed", action = "store_true",
                        help = "dashed linestyle, useful for a second snapshot")
    parser.add_argument("shpc_dir", help = "directories containing the "
                        "collections of shapefiles", nargs = "+")
    parser.add_argument("--save", metavar = "format",
                        help = "Save file to specified format")
    args = parser.parse_args()

    if args.grid or args.window is None:
        file = path.join(args.shpc_dir[0], "grid_nml.txt")
    
        try:
            grid_nml = f90nml.read(file)["grid_nml"]
        except FileNotFoundError:
            print("grid_nml.txt not found. Will not plot bounding box.")
            grid_nml = None

    if args.grid:
        width = (grid_nml["NLON"] - 1) * grid_nml["STEP_DEG"][0]
        height = (grid_nml["NLat"] - 1) * grid_nml["STEP_DEG"][1]
        longitude = np.linspace(grid_nml["CORNER_DEG"][0],
                                grid_nml["CORNER_DEG"][0] + width,
                                grid_nml["NLON"])
        latitude = np.linspace(grid_nml["CORNER_DEG"][1],
                               grid_nml["CORNER_DEG"][1] + height,
                               grid_nml["NLat"])
    
    if args.window is not None:
        llcrnrlon, llcrnrlat, urcrnrlon, urcrnrlat = args.window
        
        if urcrnrlon - llcrnrlon > 360:
            sys.exit("bad values of urcrnrlon and llcrnrlon")

        if args.grid:
            longitude += np.ceil((llcrnrlon - longitude) / 360) * 360
            # (in [llcrnrlon, llcrnrlon + 2 pi[)

            lon_mask = longitude <= urcrnrlon
            lat_mask = np.logical_and(latitude >= llcrnrlat,
                                      latitude <= urcrnrlat)
    elif args.grid:
        lon_mask = np.ones(len(longitude), dtype = bool)
        lat_mask = np.ones(len(latitude), dtype = bool)

    fig = plt.figure()
    src_crs = ccrs.PlateCarree()
    projection = ccrs.PlateCarree()
    ##projection = ccrs.NorthPolarStereo()
    ax = plt.axes(projection = projection)

    if args.grid:
        lon_2d, lat_2d = np.meshgrid(longitude[lon_mask],
                                     latitude[lat_mask])
        ax.plot(lon_2d.reshape(-1), lat_2d.reshape(-1), transform = src_crs,
                marker = "+", color = "gray", linestyle = "None")

    if args.window is None: plot_grid_bb(grid_nml, ax)

    for shpc_dir in args.shpc_dir:
        handler = util_eddies.open_shpc(shpc_dir)

        if handler["ishape_last"] is None:
            print(f"{shpc_dir}: We will use all the shapes.")
            handler["ishape_last"] = [len(handler["readers"]["extremum"]) - 1]

        if len(handler["ishape_last"]) == 1:
            if args.date is not None and args.date != handler["d_init"]:
                sys.exit(f"{shpc_dir}: Bad value of date option")
            d = handler["d_init"]
        else:
            if args.date is None:
                print(f"{shpc_dir}: No date option, plotting first date:",
                      handler["d_init"])
                d = handler["d_init"]
            else:
                d = args.date

        ishape_list = select_ishapes(d, handler, args.window)
        if len(ishape_list) == 0: print(f"{shpc_dir}: No eddy found")
        snapshot(ax, ishape_list, handler, dashed = args.dashed,
                 light = args.light)
        
    ax.set_title(f"d = {d}", y = 1.05)
    ax.gridlines(draw_labels = True)
    ax.coastlines()

    if args.save:
        plt.savefig(f"plot_eddy_contours.{args.save}")
        print(f'Created "plot_eddy_contours.{args.save}".')
    else:
        plt.show()
